package ca.carleton.gcrc.couch.onUpload.multimedia;

import java.io.File;

import org.apache.log4j.Logger;

import net.sf.json.JSONObject;

import ca.carleton.gcrc.couch.onUpload.FileConversionContext;
import ca.carleton.gcrc.couch.onUpload.FileConversionMetaData;
import ca.carleton.gcrc.couch.onUpload.FileConversionPlugin;
import ca.carleton.gcrc.couch.onUpload.MultimediaProgressAdaptor;
import ca.carleton.gcrc.couch.onUpload.UploadConstants;
import ca.carleton.gcrc.olkit.multimedia.converter.MultimediaConversionRequest;
import ca.carleton.gcrc.olkit.multimedia.converter.MultimediaConverter;
import ca.carleton.gcrc.olkit.multimedia.converter.impl.MultimediaConverterImpl;
import ca.carleton.gcrc.olkit.multimedia.file.SystemFile;
import ca.carleton.gcrc.olkit.multimedia.utils.MimeUtils;
import ca.carleton.gcrc.olkit.multimedia.utils.MimeUtils.MultimediaClass;

public class MultimediaFileConverter implements FileConversionPlugin {

	final protected Logger logger = Logger.getLogger(this.getClass());

	private MultimediaConverter mmConverter = new MultimediaConverterImpl();

	@Override
	public String getName() {
		return "Multimedia Converter";
	}

	@Override
	public boolean handlesFileClass(String fileClass) {
		if( MultimediaClass.AUDIO.getValue().equals(fileClass) ) {
			return true;
		}
		if( MultimediaClass.VIDEO.getValue().equals(fileClass) ) {
			return true;
		}
		if( MultimediaClass.IMAGE.getValue().equals(fileClass) ) {
			return true;
		}
		return false;
	}

	@Override
	public FileConversionMetaData getFileMetaData(File file) {
		
		FileConversionMetaData result = new FileConversionMetaData();
		
		try {
			SystemFile sf = SystemFile.getSystemFile(file);
			String mimeType = sf.getMimeType();
			String mimeEncoding = sf.getMimeEncoding();
	
			// Is it a known MIME type?
			MultimediaClass aClass = MimeUtils.getMultimediaClassFromMimeType(sf.getMimeType());
			if( MultimediaClass.AUDIO == aClass 
			 || MultimediaClass.VIDEO == aClass 
			 || MultimediaClass.IMAGE == aClass 
			 ) {
				String fileClass = aClass.getValue();
				
				result.setMimeType(mimeType);
				result.setMimeEncoding(mimeEncoding);
				result.setFileClass(fileClass);
				result.setFileConvertable(true);
			}
		} catch(Exception e) {
			// Ignore
		}
		
		return result;
	}

	@Override
	public void analyzeFile(FileConversionContext conversionContext) throws Exception {
		JSONObject analyzedObj = conversionContext.getAttachmentDescription();
		JSONObject originalObj = analyzedObj.optJSONObject("original");
		
		// Figure out media file located on disk
		File file = conversionContext.getOriginalFile();
		String mimeType = originalObj.getString(UploadConstants.MIME_KEY);

		// Perform conversion(s)
		MultimediaConversionRequest request = new MultimediaConversionRequest();
		request.setInFile( file );
		request.setThumbnailRequested(true);
		request.setProgress( new MultimediaProgressAdaptor() );
		
		MultimediaClass mmClass = MimeUtils.getMultimediaClassFromMimeType(mimeType);
		if( MultimediaClass.VIDEO == mmClass ) {
			// For video file, convert to appropriate file type
			mmConverter.convertVideo(request);

		} else if( MultimediaClass.AUDIO == mmClass ) {
			// For audio file, convert to appropriate file type
			mmConverter.convertAudio(request);
			
		} else if( MultimediaClass.IMAGE == mmClass ) {
			// For image file, convert to appropriate file type
			mmConverter.convertImage(request);
			
		} else {
			throw new Exception("Unknown multimedia class: "+mmClass);
		}
		
		// Report original size
		if( request.getInHeight() != 0 && request.getInWidth() != 0 ) {
			originalObj.put("height", request.getInHeight());
			originalObj.put("width", request.getInWidth());
		}

		// Report converted object
		{
			File convertedFile = request.getOutFile();
			SystemFile convertedSf = SystemFile.getSystemFile(convertedFile);

			analyzedObj.put(UploadConstants.UPLOAD_STATUS_KEY, UploadConstants.UPLOAD_STATUS_WAITING_FOR_APPROVAL);
			analyzedObj.put("conversionPerformed", request.isConversionPerformed());
			analyzedObj.put(UploadConstants.MEDIA_FILE_KEY, convertedFile.getName());
			analyzedObj.put(UploadConstants.SIZE_KEY, convertedFile.length());
			analyzedObj.put(UploadConstants.MIME_KEY, convertedSf.getMimeType());
			analyzedObj.put(UploadConstants.ENCODING_KEY, convertedSf.getMimeEncoding());
			if( request.getOutHeight() != 0 && request.getOutWidth() != 0 ) {
				analyzedObj.put("height", request.getOutHeight());
				analyzedObj.put("width", request.getOutWidth());
			}
		}

		// Report thumbnail object
		if( request.isThumbnailCreated() ) {
			File thumbFile = request.getThumbnailFile();
			SystemFile thumbSf = SystemFile.getSystemFile(thumbFile);
			
			String thumbnailAttachmentName = computeThumbnailName(analyzedObj.getString("attachmentName"));
			JSONObject thumbnailObj = conversionContext.createAttachmentDescription(thumbnailAttachmentName);

			thumbnailObj.put(UploadConstants.UPLOAD_STATUS_KEY, UploadConstants.UPLOAD_STATUS_WAITING_FOR_APPROVAL);
			thumbnailObj.put(UploadConstants.MIME_CLASS_KEY, "image");
			thumbnailObj.put("originalName", analyzedObj.opt("originalName"));
			thumbnailObj.put(UploadConstants.MEDIA_FILE_KEY, thumbFile.getName());
			thumbnailObj.put("source", analyzedObj.opt("attachmentName"));

			thumbnailObj.put(UploadConstants.SIZE_KEY, thumbFile.length());
			thumbnailObj.put(UploadConstants.MIME_KEY, thumbSf.getMimeType());
			thumbnailObj.put(UploadConstants.ENCODING_KEY, thumbSf.getMimeEncoding());

			if( request.getThumbnailHeight() != 0 && request.getThumbnailWidth() != 0 ) {
				thumbnailObj.put("height", request.getThumbnailHeight());
				thumbnailObj.put("width", request.getThumbnailWidth());
			}

			analyzedObj.put("thumbnail", thumbnailAttachmentName);
		}
	}

	@Override
	public void approveFile(FileConversionContext conversionContext) throws Exception {
		// Upload file
		String attachementName = conversionContext.getAttachmentName();
		File file = conversionContext.getFile();
		String mimeType = conversionContext.getContentType();
		conversionContext.uploadFile(attachementName, file, mimeType);
	}

	private String computeThumbnailName(String attachmentName) {
		if( null == attachmentName ) {
			return "thumbnail";
		}
		
		// Select a different file name
		String prefix = "";
		String suffix = "";
		int pos = attachmentName.indexOf('.', 1);
		if( pos < 0 ) {
			prefix = attachmentName;
		} else {
			prefix = attachmentName.substring(0, pos);
			suffix = attachmentName.substring(pos);
		}
		
		String thumbnailName = prefix + "_thumb" + suffix;
		
		return thumbnailName;
	}
}
